from typing import Any, List, Optional, Union, Dict, Type

from ovos_bus_client.client import MessageBusClient
from ovos_config import Configuration
from ovos_utils.fakebus import FakeBus

from ovos_plugin_manager.templates.pipeline import ConfidenceMatcherPipeline, PipelinePlugin
from ovos_plugin_manager.utils import PluginTypes, find_plugins, load_plugin

# Typing aliases
PipelineID = str
PipelineMatcherID = str


def find_pipeline_plugins() -> Dict[PipelineID, Type[PipelinePlugin]]:
    """
    Discover and return all installed pipeline plugins.

    Returns:
        A dictionary mapping pipeline plugin IDs to their classes.
    """
    return find_plugins(PluginTypes.PIPELINE)


def load_pipeline_plugin(module_name: str) -> Type[PipelinePlugin]:
    """
    Load a pipeline plugin class by name.

    Args:
        module_name: The name of the plugin to load.

    Returns:
        The uninstantiated plugin class.
    """
    return load_plugin(module_name, PluginTypes.PIPELINE)


class OVOSPipelineFactory:
    """
    Factory class for discovering, loading, and managing OVOS pipeline plugins.
    """

    @staticmethod
    def get_installed_pipeline_ids() -> List[PipelineID]:
        """
        List all installed pipeline plugin identifiers.

        Returns:
            A list of installed pipeline plugin IDs.
        """
        return list(find_pipeline_plugins().keys())

    @staticmethod
    def get_installed_pipeline_matcher_ids() -> List[PipelineMatcherID]:
        """
        List all available pipeline matcher identifiers, including confidence levels.

        Returns:
            A list of matcher IDs.
        """
        pipelines: List[PipelineMatcherID] = []
        for plug_id, clazz in find_pipeline_plugins().items():
            if issubclass(clazz, ConfidenceMatcherPipeline):
                pipelines.extend([
                    f"{plug_id}-low",
                    f"{plug_id}-medium",
                    f"{plug_id}-high"
                ])
            else:
                pipelines.append(plug_id)
        return pipelines

    @classmethod
    def load_plugin(
            cls,
            pipe_id: PipelineID,
            bus: Optional[Union[MessageBusClient, FakeBus]] = None,
            config: Optional[Dict[str, Any]] = None
    ) -> PipelinePlugin:
        """
        Load a pipeline plugin instance.

        Args:
            pipe_id: The pipeline plugin ID.
            bus: Optional message bus client.
            config: Optional configuration for the plugin.

        Returns:
            An instance of the loaded pipeline plugin.
        """
        config = config or Configuration().get("intents", {}).get(pipe_id, {})
        clazz = find_pipeline_plugins().get(pipe_id)
        if not clazz:
            raise ValueError(f"Unknown pipeline plugin: {pipe_id}")
        plugin_instance = clazz(bus, config)
        return plugin_instance



import json
import os
from ovos_plugin_manager.templates.tts import TTS, TTSContext, TTSValidator, TextToSpeechCache, ConcatTTS
from ovos_plugin_manager.utils import PluginTypes, PluginConfigTypes
from ovos_utils.log import LOG
from ovos_utils.xdg_utils import xdg_data_home
from hashlib import md5


def find_tts_plugins() -> dict:
    """
    Find all installed plugins
    @return: dict plugin names to entrypoints
    """
    from ovos_plugin_manager.utils import find_plugins
    return find_plugins(PluginTypes.TTS)


def load_tts_plugin(module_name: str) -> type(TTS):
    """
    Get an uninstantiated class for the requested module_name
    @param module_name: Plugin entrypoint name to load
    @return: Uninstantiated class
    """
    from ovos_plugin_manager.utils import load_plugin
    return load_plugin(module_name, PluginTypes.TTS)


def get_tts_configs() -> dict:
    """
    Get a dict of plugin names to valid TTS configuration
    @return: dict plugin name to dict of str lang to list of dict valid configs
    """
    from ovos_plugin_manager.utils.config import load_configs_for_plugin_type
    return load_configs_for_plugin_type(PluginTypes.TTS)


def get_tts_module_configs(module_name: str) -> dict:
    """
    Get a dict of lang to list of valid config dicts for a specific plugin
    @param module_name: name of plugin to get configurations for
    @return: {lang: [list of config dicts]}
    """
    from ovos_plugin_manager.utils.config import load_plugin_configs
    configs = load_plugin_configs(module_name, PluginConfigTypes.TTS)
    # let's sort by priority key
    for k, v in configs.items():
        configs[k] = sorted(v, key=lambda c: c.get("priority", 60))
    return configs


def get_tts_lang_configs(lang, include_dialects=False):
    """
    Get a dict of plugins names to sorted list of valid configurations
    @param lang: language to get configurations for (i.e. en, en-US)
    @param include_dialects: If true, include configs for other locales
        (i.e. include en-GB configs for lang=en-US)
    @return: dict plugin name to list of valid configs sorted by priority
    """
    from ovos_plugin_manager.utils.config import get_plugin_language_configs
    return get_plugin_language_configs(PluginTypes.TTS, lang, include_dialects)


def get_tts_supported_langs():
    """
    Get a dict of languages to valid configuration options
    @return: dict lang to list of plugins that support that lang
    """
    from ovos_plugin_manager.utils.config import get_plugin_supported_languages
    return get_plugin_supported_languages(PluginTypes.TTS)


def get_tts_config(config: dict = None, module: str = None) -> dict:
    """
    Get relevant configuration for factory methods
    @param config: global Configuration OR plugin class-specific configuration
    @param module: TTS module to get configuration for
    @return: plugin class-specific configuration
    """
    from ovos_plugin_manager.utils.config import get_plugin_config
    return get_plugin_config(config, 'tts', module)


def get_voice_id(plugin_name, lang, tts_config):
    tts_hash = md5(json.dumps(tts_config,
                              sort_keys=True).encode("utf-8")).hexdigest()
    return f"{plugin_name}_{lang}_{tts_hash}"


def scan_voices():
    voice_ids = {}
    for lang in get_tts_supported_langs():
        VOICES_FOLDER = f"{xdg_data_home()}/OPM/voice_configs/{lang}"
        os.makedirs(VOICES_FOLDER, exist_ok=True)
        for plug, voices in get_tts_lang_configs(lang,
                                                 include_dialects=True).items():
            for voice in voices:
                voiceid = get_voice_id(plug, lang, voice)
                if "meta" not in voice:
                    voice["meta"] = {}
                noise = ["priority", "display_name", "offline", "gender"]
                for k in noise:
                    if k in voice:
                        voice["meta"][k] = voice.pop(k)
                voice["module"] = plug
                with open(f"{VOICES_FOLDER}/{voiceid}.json", "w") as f:
                    json.dump(voice, f, indent=4, ensure_ascii=False)
                voice_ids[voiceid] = voice
    return voice_ids


def get_voices(scan=False):
    if scan:
        scan_voices()
    voice_ids = {}
    for lang in get_tts_supported_langs():
        VOICES_FOLDER = f"{xdg_data_home()}/OPM/voice_configs/{lang}"
        for voice in os.listdir(VOICES_FOLDER):
            with open(f"{VOICES_FOLDER}/{voice}") as f:
                voice_ids[voice] = json.load(f)
    return voice_ids


class OVOSTTSFactory:
    """ replicates the base mycroft class, but uses only OPM enabled plugins"""

    @staticmethod
    def get_class(config=None):
        """Factory method to get a TTS engine class based on configuration.

        The configuration file ``mycroft.conf`` contains a ``tts`` section with
        the name of a TTS module to be read by this method.

        "tts": {
            "module": <engine_name>
        }
        """
        config = config or get_tts_config()
        tts_module = config.get("module") or "ovos-tts-plugin-dummy"
        return load_tts_plugin(tts_module)

    @staticmethod
    def create(config=None):
        """Factory method to create a TTS engine based on configuration.

        The configuration file ``mycroft.conf`` contains a ``tts`` section with
        the name of a TTS module to be read by this method.

        "tts": {
            "module": <engine_name>
        }
        """
        tts_config = get_tts_config(config)
        tts_module = tts_config.get('module')
        if not tts_module:
            raise ValueError("tts 'module' is not set in config")
        try:
            clazz = OVOSTTSFactory.get_class(tts_config)
            if clazz:
                LOG.info(f'Found plugin {tts_module}')
                tts = clazz(config=tts_config)
                tts._plugin_id = tts_module
                tts.validator.validate()
                LOG.info(f'Loaded plugin {tts_module}')
            else:
                raise RuntimeError(f"unknown plugin: {tts_module}")
        except Exception:
            plugins = find_tts_plugins()
            modules = ",".join(plugins.keys())
            LOG.exception(f'The TTS plugin "{tts_module}" could not be loaded.'
                          f'\nAvailable modules: {modules}')
            raise
        return tts



from ovos_plugin_manager.utils import PluginTypes, PluginConfigTypes
from ovos_config import Configuration
from ovos_utils.log import LOG
from ovos_plugin_manager.templates.vad import VADEngine


def find_vad_plugins() -> dict:
    """
    Find all installed plugins
    @return: dict plugin names to entrypoints
    """
    from ovos_plugin_manager.utils import find_plugins
    return find_plugins(PluginTypes.VAD)


def load_vad_plugin(module_name: str) -> type(VADEngine):
    """
    Get an uninstantiated class for the requested module_name
    @param module_name: Plugin entrypoint name to load
    @return: Uninstantiated class
    """
    from ovos_plugin_manager.utils import load_plugin
    return load_plugin(module_name, PluginTypes.VAD)


def get_vad_configs() -> dict:
    """
    Get valid plugin configurations by plugin name
    @return: dict plugin names to list of dict configurations
    """
    from ovos_plugin_manager.utils.config import load_configs_for_plugin_type
    return load_configs_for_plugin_type(PluginTypes.VAD)


def get_vad_module_configs(module_name: str) -> dict:
    """
    Get valid configurations for the specified plugin
    @param module_name: plugin to get configuration for
    @return: dict configurations by language (if provided)
    """
    # VAD plugins return [list of config dicts] or {module_name: [list of config dicts]}
    from ovos_plugin_manager.utils.config import load_plugin_configs
    cfgs = load_plugin_configs(module_name,
                               PluginConfigTypes.VAD)
    return {module_name: cfgs} if isinstance(cfgs, list) else cfgs


def get_vad_config(config: dict = None) -> dict:
    """
    Get relevant configuration for factory methods
    @param config: global Configuration OR plugin class-specific configuration
    @return: plugin class-specific configuration
    """
    from ovos_plugin_manager.utils.config import get_plugin_config
    config = config or Configuration()
    if "listener" in config and "VAD" not in config:
        config = get_plugin_config(config, "listener")
    if "VAD" in config:
        config = get_plugin_config(config, "VAD")
    return config


class OVOSVADFactory:

    @staticmethod
    def get_class(config=None):
        """Factory method to get a VAD engine class based on configuration.

        The configuration file ``mycroft.conf`` contains a ``vad`` section with
        the name of a VAD module to be read by this method.

        "VAD": {
            "module": <engine_name>
        }
        """
        config = get_vad_config(config)
        vad_module = config.get("module")
        if not vad_module:
            raise ValueError(f"VAD Plugin not configured in: {config}")
        if vad_module == "dummy":
            return VADEngine
        return load_vad_plugin(vad_module)

    @classmethod
    def create(cls, config=None):
        """Factory method to create a VAD engine based on configuration.

        The configuration file ``mycroft.conf`` contains a ``VAD`` section with
        the name of a VAD module to be read by this method.

        "VAD": {
            "module": <engine_name>
        }
        """
        config = config or Configuration()
        if "listener" in config:
            config = config["listener"]
        if "VAD" in config:
            config = config["VAD"]
        plugin = config.get("module")
        if not plugin:
            raise ValueError(f"VAD Plugin not configured in: {config}")

        plugin_config = config.get(plugin, {})
        fallback = plugin_config.get("fallback_module")

        try:
            clazz = OVOSVADFactory.get_class(config)
            return clazz(plugin_config)
        except Exception:
            LOG.exception(f'VAD plugin {plugin} could not be loaded!')
            if fallback in config and fallback != plugin:
                LOG.info(f"Attempting to load fallback plugin instead: {fallback}")
                config["module"] = fallback
                return cls.create(config)
            raise
